//	GamePlayState.cpp - The main gameplay state.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "MinimalSounder.h"
#include "drawing/Drawer.h"
#include "GamePlayState.h"
#include "ColourHelpers.h"
#include "DebugHeaders.h"
#include "MiscHelper.h"
#include "Random.h"
#include "Ball.h"

#include "CaptureTheFlagPlayer.h"
#include "FootballPlayer.h"
#include "SnakePlayer.h"
#include "PongPlayer.h"
#include "DeadPlayer.h"

#include "Zombie.h"
#include "Bullet.h"
#include "Apple.h"
#include "Flag.h"

#include <sstream>

//------------------------------------------------------------------------------
GameStateCreator<GamePlayState> GamePlayState::creator;

//------------------------------------------------------------------------------
GamePlayState::GamePlayState(Application& app, SounderPtr& sound):
GameState(app, sound),
shakeVal(0.0f),
shakeIndex(0.0f),
coloursSet(false),
leftType(Pong),
rightType(Pong),
leftGoals(false),
rightGoals(false),
backgroundColour(0.0f),
backgroundTween(backgroundColour, backgroundColour, 0.01f),
wallColour(1.0f),
wallTween(wallColour, wallColour, 0.01f),
noiseSeedInc(0.0f),
noiseCount(2),
satCon(0.0f),
saturation(0.0f),
sounder(dynamic_pointer_cast<MinimalSounder>(sound))
{
	const TwoFloats& drawSize(getDrawingAreaSize());

	backgroundTween.setIndex(1.0f);
	wallTween.setIndex(1.0f);

	gameCounter.registerNotification(L"LeftScore", 1, this);
	gameCounter.registerNotification(L"RightScore", 1, this);

	players[0] = shared_ptr<Player>(new PongPlayer(TwoFloats(16.0f,
															 drawSize.y*0.5f)));
	players[1] = shared_ptr<Player>(new PongPlayer(TwoFloats((drawSize.x-16.0f),
															 drawSize.y*0.5f)));
	firing[0] = 0;
	firing[1] = 0;
	zombieCount[0] = 0;
	zombieCount[1] = 0;

	ball = shared_ptr<Ball>(new Ball(drawSize*0.5f,
									 (Random::getGlobal().getRandomFloat() < 0.5f),
									 sounder,
									 particles,
									 *this));

	flagBases[0].setPosition(TwoFloats(64.0f, drawSize.y*0.5f), -0.01f);
	flagBases[1].setPosition(TwoFloats((drawSize.x-64.0f), drawSize.y*0.5f),
							 -0.01f);

	upIsDown[0] = false;
	downIsDown[0] = false;
	leftIsDown[0] = false;
	rightIsDown[0] = false;
	upIsDown[1] = false;
	downIsDown[1] = false;
	leftIsDown[1] = false;
	rightIsDown[1] = false;
}

//------------------------------------------------------------------------------
GamePlayState::~GamePlayState()
{

}

//------------------------------------------------------------------------------
void GamePlayState::update()
{
	int i;
	TwoFloats tempPos;
	TwoFloats tempDir;
	const TwoFloats& drawSize(getDrawingAreaSize());

	if(!backgroundTween.getFinished())
	{
		backgroundTween.tick();
		if(backgroundTween.getFinished())
		{
			ThreeFloats tempCol(0.0f, 0.4f, 0.8f);

			tempCol.x = Random::getGlobal().getRandomFloat();
			tempCol = ColourHelpers::hslToRgb(tempCol);
			backgroundTween.reset(tempCol);
		}
	}

	if(!wallTween.getFinished())
	{
		wallTween.tick();
		if(wallTween.getFinished())
		{
			ThreeFloats tempCol(0.0f, 0.66f, 0.7f);

			tempCol.x = Random::getGlobal().getRandomFloat();
			tempCol = ColourHelpers::hslToRgb(tempCol);
			wallTween.reset(tempCol);
		}
	}

	//Generate new zombies.
	if(zombieCount[0] > 0)
	{
		--zombieCount[0];
		if(zombieCount[0] == 1)
		{
			if(npcs.size() < 256)
			{
				tempPos.x = -24.0f;
				tempPos.y = Random::getGlobal().getRandomFloat() * drawSize.y;

				npcs.push_back(shared_ptr<NPC>(new Zombie(tempPos,
														  players[0]->getColour(),
														  players[1],
														  gameCounter)));
			}

			zombieCount[0] = 5;
		}
	}
	if(zombieCount[1] > 0)
	{
		--zombieCount[1];
		if(zombieCount[1] == 1)
		{
			if(npcs.size() < 256)
			{
				tempPos.x = drawSize.x + 24.0f;
				tempPos.y = Random::getGlobal().getRandomFloat() * drawSize.y;

				npcs.push_back(shared_ptr<NPC>(new Zombie(tempPos,
														  players[1]->getColour(),
														  players[0],
														  gameCounter)));
			}

			zombieCount[1] = 5;
		}
	}

	//Update players.
	players[0]->update(drawSize);
	players[1]->update(drawSize);

	//Fire bullets.
	if(firing[0] > 0)
	{
		--firing[0];
		if(firing[0] == 1)
		{
			firing[0] = 17;

			if(lastJoy[0] == 0.0f)
				tempDir = players[0]->getDirection();
			else
				tempDir = lastJoy[0].normal();
			npcs.push_back(shared_ptr<NPC>(new Bullet(players[0]->getPosition(),
													  players[0]->getColour(),
													  tempDir)));

			sounder->triggerShoot(Random::getGlobal().getRandomInt(8));
		}
	}
	if(firing[1] > 0)
	{
		--firing[1];
		if(firing[1] == 1)
		{
			firing[1] = 17;

			if(lastJoy[1] == 0.0f)
				tempDir = players[1]->getDirection();
			else
				tempDir = lastJoy[1].normal();
			npcs.push_back(shared_ptr<NPC>(new Bullet(players[1]->getPosition(),
													  players[1]->getColour(),
													  tempDir)));

			sounder->triggerShoot(Random::getGlobal().getRandomInt(8));
		}
	}

	//Update ball.
	if(ball->getOwner().x == 0.0f)
		ball->update(drawSize);
	else
	{
		if(ball->getOwner() == players[0]->getColour())
		{
			if(players[0]->getDirection() != 0.0f)
			{
				tempPos = players[0]->getPosition();
				tempPos += players[0]->getDirection() * (players[0]->getRadius() + ball->getRadius());
				ball->setPosition(tempPos);
				//ball->setDirection(players[0]->getDirection());
			}
			else
			{
				ball->setOwner(0.0f);
				ball->resetSpeed(Random::getGlobal().getRandomFloat() < 0.5f);

				tempDir = ball->getPosition();
				tempDir = (tempDir - players[0]->getPosition()).normal();

				tempPos = ball->getPosition() + (tempDir * (ball->getRadius()+players[0]->getRadius()));
				ball->setPosition(tempPos);
				//ball->setDirection(tempDir);
			}
		}
		else
		{
			if(players[1]->getDirection() != 0.0f)
			{
				tempPos = players[1]->getPosition();
				tempPos += players[1]->getDirection() * (players[1]->getRadius() + ball->getRadius());
				ball->setPosition(tempPos);
				//ball->setDirection(players[1]->getDirection());
			}
			else
			{
				ball->setOwner(0.0f);
				ball->resetSpeed(Random::getGlobal().getRandomFloat() < 0.5f);

				tempDir = ball->getPosition();
				tempDir = (tempDir - players[0]->getPosition()).normal();

				tempPos = ball->getPosition() + (tempDir * (ball->getRadius()+players[0]->getRadius()));
				ball->setPosition(tempPos);
				//ball->setDirection(tempDir);
			}
		}
	}

	//Update colours.
	{
		starColour = backgroundColour;

		starColour = ColourHelpers::rgbToHsl(starColour);
		appleColour = starColour;
		starColour.x += 0.33f;
		starColour.y *= 0.5f;
		starColour.z *= 0.7f;
		if(starColour.x > 1.0f)
			starColour.x -= 1.0f;
		appleColour.x += 0.66f;
		if(appleColour.x > 1.0f)
			appleColour.x -= 1.0f;
		starColour = ColourHelpers::hslToRgb(starColour);
		appleColour = ColourHelpers::hslToRgb(appleColour);
	}

	//Update NPCs.
	for(i=0;i<(int)npcs.size();++i)
	{
		if(npcs[i]->getName() == L"Apple")
		{
			shared_ptr<Apple> apple(dynamic_pointer_cast<Apple>(npcs[i]));

			apple->setColour(appleColour);
			apple->setBackColour(starColour);
		}

		npcs[i]->update(drawSize);
	}

	//Collision detection.
	collisionDetection();

	//Remove any dead NPCs.
	for(i=(npcs.size()-1);i>=0;--i)
	{
		if(npcs[i]->isDead())
		{
			if((npcs[i]->getName() == L"Apple") &&
			   ((leftType == Snake) || (rightType == Snake)))
			{
				addApple();
			}

			particles.trigger(npcs[i]->getPosition(), npcs[i]->getColour());
			shakeScreen();
			npcs.erase(npcs.begin()+i);
		}
	}

	//Test for dead players.
	if(leftType == ZombieApocalypse)
	{
		shared_ptr<CaptureTheFlagPlayer> p(dynamic_pointer_cast<CaptureTheFlagPlayer>(players[0]));

		if(p)
		{
			if(p->getDead())
			{
				++leftType;
				players[0] = shared_ptr<Player>(new DeadPlayer(players[0]->getPosition(),
															   players[0]->getColour()));

				zombieCount[0] = 0;
				zombieCount[1] = 0;

				players[1]->keyUp(L"P1Up");
				players[1]->keyUp(L"P1Down");
				players[1]->keyUp(L"P1Left");
				players[1]->keyUp(L"P1Right");

				if(gameOverColour == 0.0f)
				{
					if(gameCounter.getCount(L"LeftScore") > gameCounter.getCount(L"RightScore"))
					{
						gameOverString = L"Red wins (posthumously)";
						gameOverColour = players[0]->getColour() * 0.75f;
					}
					else
					{
						gameOverString = L"Blue wins (posthumously)";
						gameOverColour = players[1]->getColour() * 0.75f;
					}

					sounder->triggerGameOver(Random::getGlobal().getRandomInt(8));
				}
			}
		}
	}
	if(rightType == ZombieApocalypse)
	{
		shared_ptr<CaptureTheFlagPlayer> p(dynamic_pointer_cast<CaptureTheFlagPlayer>(players[1]));

		if(p)
		{
			if(p->getDead())
			{
				++rightType;
				players[1] = shared_ptr<Player>(new DeadPlayer(players[1]->getPosition(),
															   players[1]->getColour()));

				zombieCount[0] = 0;
				zombieCount[1] = 0;

				players[0]->keyUp(L"P1Up");
				players[0]->keyUp(L"P1Down");
				players[0]->keyUp(L"P1Left");
				players[0]->keyUp(L"P1Right");

				if(gameOverColour == 0.0f)
				{
					if(gameCounter.getCount(L"LeftScore") > gameCounter.getCount(L"RightScore"))
					{
						gameOverString = L"Red wins (posthumously)";
						gameOverColour = players[0]->getColour() * 0.75f;
					}
					else
					{
						gameOverString = L"Blue wins (posthumously)";
						gameOverColour = players[1]->getColour() * 0.75f;
					}
				}
			}
		}
	}

	//Score detection.
	if((leftType < Dead) && (rightType < Dead))
	{
		const TwoFloats& ballPos(ball->getPosition());
		const TwoFloats leftBase(64.0f, drawSize.y*0.5f);
		const TwoFloats rightBase(drawSize.x-64.0f, drawSize.y*0.5f);

		if((ballPos.x < -16.0f) && (rightType < CaptureTheFlag))
			handleScore(true);
		else if((ballPos.x > (drawSize.x+16.0f)) && (leftType < CaptureTheFlag))
			handleScore(false);
		else if((ballPos.x < -16.0f) || (ballPos.x > (drawSize.x+16.0f)))
		{
			ball->setPosition(drawSize*0.5f);
			ball->setOwner(0.0f);
			ball->resetSpeed(Random::getGlobal().getRandomFloat() < 0.5f);
		}

		if(players[0]->getFlag())
		{
			if(players[0]->getPosition().getDistance(leftBase) < (players[0]->getRadius() * 2.0f))
			{
				for(i=0;i<(int)npcs.size();++i)
				{
					if(npcs[i]->getName() == L"Flag")
					{
						shared_ptr<Flag> flag(dynamic_pointer_cast<Flag>(npcs[i]));

						if(flag->getColour() != players[0]->getColour())
						{
							shared_ptr<Flag> nullFlag;
							shared_ptr<CaptureTheFlagPlayer> p(dynamic_pointer_cast<CaptureTheFlagPlayer>(players[0]));

							p->setFlag(nullFlag);
							npcs.erase(npcs.begin()+i);

							handleScore(false);

							break;
						}
					}
				}
			}
		}
		if(players[1]->getFlag())
		{
			if(players[1]->getPosition().getDistance(rightBase) < (players[1]->getRadius() * 2.0f))
			{
				for(i=0;i<(int)npcs.size();++i)
				{
					if(npcs[i]->getName() == L"Flag")
					{
						shared_ptr<Flag> flag(dynamic_pointer_cast<Flag>(npcs[i]));

						if(flag->getColour() != players[1]->getColour())
						{
							shared_ptr<Flag> nullFlag;
							shared_ptr<CaptureTheFlagPlayer> p(dynamic_pointer_cast<CaptureTheFlagPlayer>(players[1]));

							p->setFlag(nullFlag);
							npcs.erase(npcs.begin()+i);

							handleScore(true);

							break;
						}
					}
				}
			}
		}
	}

	if((leftType == CaptureTheFlag) || (rightType == CaptureTheFlag))
	{
		flagBases[0].update();
		flagBases[1].update();
	}

	particles.update();

	//Update screen shake.
	if(shakeVal > 0.0f)
	{
		shakeIndex += 0.15f;
		if(shakeIndex > 1.0f)
			shakeIndex -= 1.0f;

		shakeVal -= 0.05f;
		if(shakeVal < 0.0f)
			shakeVal = 0.0f;
	}

	//Update satCon.
	if(satCon > 0.0f)
	{
		satCon -= 0.02f;
		if(satCon < 0.0f)
			satCon = 0.0f;
	}
}

//------------------------------------------------------------------------------
void GamePlayState::draw(DrawerPtr& d, float timestepInterp)
{
	unsigned int i;
	std::wstringstream tempstr;
	const TwoFloats& drawSize(d->getDrawingAreaSize());
	float shake(sinf(shakeIndex * 2.0f * NIALL_PI) * (shakeVal < 1.0f ? shakeVal : 1.0f) * 4.0f);

	d->startFrameBufferDrawing();

	//Draw background.
	d->fillRect(0.0f, drawSize, backgroundColour);

	//Shake screen.
	if(shakeVal > 0.0f)
		d->translate(TwoFloats(shake, 0.0f));

	//Draw flag bases.
	if((leftType == CaptureTheFlag) || (rightType == CaptureTheFlag))
	{
		flagBases[0].draw(d, starColour);
		flagBases[1].draw(d, starColour);
	}

	//Draw NPCs.
	for(i=0;i<npcs.size();++i)
	{
		if(npcs[i]->getName() != L"Flag")
			npcs[i]->draw(d);
	}

	//Draw players.
	players[0]->draw(d);
	players[1]->draw(d);

	//Draw NPCs.
	for(i=0;i<npcs.size();++i)
	{
		if(npcs[i]->getName() == L"Flag")
			npcs[i]->draw(d);
	}

	//Draw ball.
	ball->draw(d);

	particles.draw(d, timestepInterp);

	//Draw walls.
	d->fillRect(0.0f, TwoFloats(drawSize.x, 32.0f), wallColour);
	d->fillRect(TwoFloats(0.0f, drawSize.y-32.0f),
				TwoFloats(drawSize.x, 32.0f),
				wallColour);
	if(leftGoals)
	{
		d->fillRect(0.0f,
					TwoFloats(32.0f, (drawSize.y*0.5f)-128.0f),
					wallColour);
		d->fillRect(TwoFloats(0.0f, (drawSize.y*0.5f)+128.0f),
					TwoFloats(32.0f, (drawSize.y*0.5f)-128.0f),
					wallColour);
	}
	if(rightGoals)
	{
		d->fillRect(TwoFloats(drawSize.x-32.0f, 0.0f),
					TwoFloats(32.0f, (drawSize.y*0.5f)-128.0f),
					wallColour);
		d->fillRect(TwoFloats(drawSize.x-32.0f, (drawSize.y*0.5f)+128.0f),
					TwoFloats(32.0f, (drawSize.y*0.5f)-128.0f),
					wallColour);
	}

	//Draw score.
	tempstr << gameCounter.getCount(L"LeftScore");
	d->drawText(tempstr.str(),
				L"ScoreFont",
				TwoFloats(0.0f, 160.0f),
				((d->getDrawingAreaSize().x*0.5f)-32.0f),
				TextLayout::Right,
				1.0f,
				players[0]->getColour());
	tempstr.clear();
	tempstr.str(L"");
	tempstr << gameCounter.getCount(L"RightScore");
	d->drawText(tempstr.str(),
				L"ScoreFont",
				TwoFloats(((d->getDrawingAreaSize().x*0.5f)+32.0f), 160.0f),
				((d->getDrawingAreaSize().x*0.5f)-32.0f),
				TextLayout::Left,
				1.0f,
				players[1]->getColour());

	//Draw game over text.
	if(gameOverString != L"")
	{
		d->drawText(gameOverString,
					L"ScoreFont",
					gameOverText,
					TwoFloats(0.0f, ((drawSize.y*0.5f)+64.0f)),
					drawSize.x,
					TextLayout::Centre,
					1.0f,
					gameOverColour);
		d->drawText(L"Play suspended indefinitely due to zombie apocalypse.",
					L"TickerFont",
					tickerText,
					TwoFloats(0.0f, drawSize.y-64.0f),
					drawSize.x,
					TextLayout::Centre);
	}

	if(shakeVal > 0.0f)
		d->translate(TwoFloats(-shake, 0.0f));

	d->endFrameBufferDrawing();

	d->useShader(L"AddNoise");
	{
		d->setShaderParam(L"noiseSeedInc", noiseSeedInc);
		//d->setShaderParam(L"brightness", brightness);
		d->setShaderParam(L"saturation", (satCon/7.0f) + saturation);
		d->setShaderParam(L"contrast", (satCon/3.0f) + saturation);

		d->drawImage(d->getFrameBufferImage(), 0.0f, 1.0f);
	}
	d->useShader(L"");

	--noiseCount;
	if(!noiseCount)
	{
		noiseCount = 8;

		noiseSeedInc += 0.01f;
		if(noiseSeedInc > 1.0f)
			noiseSeedInc -= 1.0f;
	}
}

//------------------------------------------------------------------------------
void GamePlayState::keyDown(const std::wstring& action,
							const std::wstring& keyPressed,
							int keyModifier)
{
	bool dead = ((leftType == Dead) || (rightType == Dead));

	if(action == L"P1Up")
		upIsDown[0] = true;
	else if(action == L"P1Down")
		downIsDown[0] = true;
	else if(action == L"P1Left")
		leftIsDown[0] = true;
	else if(action == L"P1Right")
		rightIsDown[0] = true;
	else if(action == L"P2Up")
		upIsDown[1] = true;
	else if(action == L"P2Down")
		downIsDown[1] = true;
	else if(action == L"P2Left")
		leftIsDown[1] = true;
	else if(action == L"P2Right")
		rightIsDown[1] = true;

	if(action == L"Escape")
		nextState(L"TitleScreen");
	else if(action != L"")
	{
		if(!dead)
		{
			if(action == L"P1Fire")
			{
				if(leftType == Football)
				{
					if(ball->getOwner() == ThreeFloats(1.0f, 0.45f, 0.45f))
					{
						TwoFloats tempDir;

						if(players[0]->getDirection() != 0.0f)
							ball->setDirection(players[0]->getDirection());
						else
						{
							tempDir = ball->getPosition();
							tempDir = (tempDir - players[0]->getPosition()).normal();
							ball->setDirection(tempDir);
						}

						ball->kick();
						sounder->triggerShoot(Random::getGlobal().getRandomInt(8));
					}
				}
				else if(leftType >= CaptureTheFlag)
					firing[0] = 2;
			}
			else if(action == L"P2Fire")
			{
				if(rightType == Football)
				{
					if(ball->getOwner() == ThreeFloats(0.53f, 0.45f, 1.0f))
					{
						TwoFloats tempDir;

						if(players[1]->getDirection() != 0.0f)
							ball->setDirection(players[1]->getDirection());
						else
						{
							tempDir = ball->getPosition();
							tempDir = (tempDir - players[1]->getPosition()).normal();
							ball->setDirection(tempDir);
						}

						ball->kick();
						sounder->triggerShoot(Random::getGlobal().getRandomInt(8));
					}
				}
				else if(rightType >= CaptureTheFlag)
					firing[1] = 2;
			}
			else if(action[1] == L'1')
			{
				players[0]->keyDown(action.substr(2));
			}
			else if(action[1] == L'2')
			{
				players[1]->keyDown(action.substr(2));
			}
		}
		else if((action == L"P1Fire") || (action == L"P2Fire"))
			nextState(L"TitleScreen");
		else if(action == L"Restart")
			nextState(L"GamePlayState");
	}
}

//------------------------------------------------------------------------------
void GamePlayState::keyUp(const std::wstring& action,
						  const std::wstring& keyPressed,
						  int keyModifier)
{
	bool dead = ((leftType == Dead) || (rightType == Dead));

	if(action == L"P1Up")
		upIsDown[0] = false;
	else if(action == L"P1Down")
		downIsDown[0] = false;
	else if(action == L"P1Left")
		leftIsDown[0] = false;
	else if(action == L"P1Right")
		rightIsDown[0] = false;
	else if(action == L"P2Up")
		upIsDown[1] = false;
	else if(action == L"P2Down")
		downIsDown[1] = false;
	else if(action == L"P2Left")
		leftIsDown[1] = false;
	else if(action == L"P2Right")
		rightIsDown[1] = false;

	if(action == L"")
		return;
	else if(!dead)
	{
		if(action == L"P1Fire")
		{
			if(leftType >= CaptureTheFlag)
				firing[0] = 0;
		}
		else if(action == L"P2Fire")
		{
			if(rightType >= CaptureTheFlag)
				firing[1] = 0;
		}
		else if(action[1] == L'1')
		{
			players[0]->keyUp(action.substr(2));
		}
		else if(action[1] == L'2')
		{
			players[1]->keyUp(action.substr(2));
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::joyMove(const std::wstring& action,
						    float value,
						    int joystick,
						    int axis)
{
	bool dead = ((leftType == Dead) || (rightType == Dead));

	if((action == L"P1Up") || (action == L"P1Down"))
		joyLast[0].y = value;
	else if((action == L"P1Left") || (action == L"P1Right"))
		joyLast[0].x = value;
	else if((action == L"P2Up") || (action == L"P2Down"))
		joyLast[1].y = value;
	else if((action == L"P2Left") || (action == L"P2Right"))
		joyLast[1].x = value;

	if(action == L"")
		return;
	else if(!dead)
	{
		if(action[1] == L'1')
		{
			std::wstring tempstr(action.substr(2));

			if(tempstr.substr(0, 4) == L"Fire")
			{
				if(leftType >= CaptureTheFlag)
				{
					std::wstring tempstr2(tempstr.substr(4));

					if((tempstr.substr(4) == L"Up") || (tempstr.substr(4) == L"Down"))
					{
						if(fabsf(value) > 0.15f)
							lastJoy[0].y = value;
						else
							lastJoy[0].y = 0.0f;
					}
					else if((tempstr.substr(4) == L"Left") || (tempstr.substr(4) == L"Right"))
					{
						if(fabsf(value) > 0.15f)
							lastJoy[0].x = value;
						else
							lastJoy[0].x = 0.0f;
					}

					if(!firing[0] && (lastJoy[0] != 0.0f))
						firing[0] = 2;
					else if(lastJoy[0] == 0.0f)
						firing[0] = 0;
				}
			}
			else
				players[0]->joystick(tempstr, value);
		}
		else if(action[1] == L'2')
		{
			std::wstring tempstr(action.substr(2));

			if(tempstr.substr(0, 4) == L"Fire")
			{
				if(rightType >= CaptureTheFlag)
				{
					std::wstring tempstr2(tempstr.substr(4));

					if((tempstr.substr(4) == L"Up") || (tempstr.substr(4) == L"Down"))
					{
						if(fabsf(value) > 0.15f)
							lastJoy[1].y = value;
						else
							lastJoy[1].y = 0.0f;
					}
					else if((tempstr.substr(4) == L"Left") || (tempstr.substr(4) == L"Right"))
					{
						if(fabsf(value) > 0.15f)
							lastJoy[1].x = value;
						else
							lastJoy[1].x = 0.0f;
					}

					if(!firing[1] && (lastJoy[1] != 0.0f))
						firing[1] = 2;
					else if(lastJoy[1] == 0.0f)
						firing[1] = 0;
				}
			}
			else
				players[1]->joystick(tempstr, value);
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::countReached(const std::wstring& counter, int value)
{
	satCon += 0.5f;
	sounder->triggerScore(Random::getGlobal().getRandomInt(8));
}

//------------------------------------------------------------------------------
void GamePlayState::collisionDetection()
{
	unsigned int i, j;
	const TwoFloats& drawSize(getDrawingAreaSize());

	//Player-ball & player-player collisions.
	players[0]->collide(ball);
	players[1]->collide(ball);
	players[0]->collide(players[1]);
	players[1]->collide(players[0]);

	//Player-NPC, NPC-ball collisions.
	for(i=0;i<npcs.size();++i)
	{
		players[0]->collide(npcs[i]);
		players[1]->collide(npcs[i]);
		npcs[i]->collide(ball);

		if(npcs[i]->getName() == L"Zombie")
		{
			for(j=0;j<npcs.size();++j)
			{
				if(npcs[j]->getName() != L"Zombie")
					npcs[i]->collide(npcs[j]);
			}
		}
	}

	//Ball-NPC collisions.

	//Ball-wall collisions.
	{
		TwoFloats ballDir(ball->getDirection());
		const TwoFloats& ballPos(ball->getPosition());

		if(ballPos.y < (32.0f+16.0f))
		{
			if(ballDir.y < 0.0f)
			{
				ballDir.y = -ballDir.y;
				//sounder->triggerCollision(Random::getGlobal().getRandomInt(8));
			}
			ball->setDirection(ballDir);
		}
		else if(ballPos.y > (drawSize.y-(32.0f+16.0f)))
		{
			if(ballDir.y > 0.0f)
			{
				ballDir.y = -ballDir.y;
				//sounder->triggerCollision(Random::getGlobal().getRandomInt(8));
			}
			ball->setDirection(ballDir);
		}

		if(leftGoals && (ballPos.x < (32.0f+16.0f)))
		{
			if((ballPos.y < ((drawSize.y*0.5f)-128.0f)) ||
			   (ballPos.y > ((drawSize.y*0.5f)+128.0f)))
			{
				ballDir.x = -ballDir.x;
				//sounder->triggerCollision(Random::getGlobal().getRandomInt(8));
			}
			ball->setDirection(ballDir);
		}
		else if(rightGoals && (ballPos.x > (drawSize.x-(32.0f+16.0f))))
		{
			if((ballPos.y < ((drawSize.y*0.5f)-128.0f)) ||
			   (ballPos.y > ((drawSize.y*0.5f)+128.0f)))
			{
				ballDir.x = -ballDir.x;
				//sounder->triggerCollision(Random::getGlobal().getRandomInt(8));
			}
			ball->setDirection(ballDir);
		}
	}

	//Bullet-wall collisions.
	for(i=0;i<npcs.size();++i)
	{
		if(npcs[i]->getName() == L"Bullet")
		{
			const TwoFloats& bulletPos(npcs[i]->getPosition());
			shared_ptr<Bullet> bullet(dynamic_pointer_cast<Bullet>(npcs[i]));

			if(bulletPos.y < (32.0f+16.0f))
			{
				bullet->setDead();
				sounder->triggerCollision(Random::getGlobal().getRandomInt(8));
			}
			else if(bulletPos.y > (drawSize.y-(32.0f+16.0f)))
			{
				bullet->setDead();
				sounder->triggerCollision(Random::getGlobal().getRandomInt(8));
			}

			if(leftGoals && (bulletPos.x < (32.0f+16.0f)))
			{
				if((bulletPos.y < ((drawSize.y*0.5f)-128.0f)) ||
				   (bulletPos.y > ((drawSize.y*0.5f)+128.0f)))
				{
					bullet->setDead();
					sounder->triggerCollision(Random::getGlobal().getRandomInt(8));
				}
			}
			else if(rightGoals && (bulletPos.x > (drawSize.x-(32.0f+16.0f))))
			{
				if((bulletPos.y < ((drawSize.y*0.5f)-128.0f)) ||
				   (bulletPos.y > ((drawSize.y*0.5f)+128.0f)))
				{
					bullet->setDead();
					sounder->triggerCollision(Random::getGlobal().getRandomInt(8));
				}
			}
		}
	}

	//Player-goal walls collisions.
	if(leftGoals)
	{
		for(i=0;i<2;++i)
		{
			TwoFloats playerPos(players[i]->getPosition());

			if(playerPos.x < (32.0f+16.0f))
			{
				shared_ptr<PongPlayer> p(dynamic_pointer_cast<PongPlayer>(players[i]));

				if(p)
				{
					if(playerPos.y < (((drawSize.y*0.5f)-128.0f)+64.0f))
					{
						playerPos.y = (((drawSize.y*0.5f)-128.0f)+64.0f);
						players[i]->setPosition(playerPos);
					}
					else if(playerPos.y > (((drawSize.y*0.5f)+128.0f)-64.0f))
					{
						playerPos.y = (((drawSize.y*0.5f)+128.0f)-64.0f);
						players[i]->setPosition(playerPos);
					}
				}
				else if((playerPos.y < ((drawSize.y*0.5f)-128.0f)) ||
					    (playerPos.y > ((drawSize.y*0.5f)+128.0f)))
				{

					playerPos.x = 32.0f+16.0f;
					players[i]->setPosition(playerPos);
				}
			}
		}
	}
	if(rightGoals)
	{
		for(i=0;i<2;++i)
		{
			TwoFloats playerPos(players[i]->getPosition());

			if(playerPos.x > (drawSize.x-(32.0f+16.0f)))
			{
				shared_ptr<PongPlayer> p(dynamic_pointer_cast<PongPlayer>(players[i]));

				if(p)
				{
					if(playerPos.y < (((drawSize.y*0.5f)-128.0f)+64.0f))
					{
						playerPos.y = (((drawSize.y*0.5f)-128.0f)+64.0f);
						players[i]->setPosition(playerPos);
					}
					else if(playerPos.y > (((drawSize.y*0.5f)+128.0f)-64.0f))
					{
						playerPos.y = (((drawSize.y*0.5f)+128.0f)-64.0f);
						players[i]->setPosition(playerPos);
					}
				}
				else if((playerPos.y < ((drawSize.y*0.5f)-128.0f)) ||
					    (playerPos.y > ((drawSize.y*0.5f)+128.0f)))
				{

					playerPos.x = (drawSize.x-(32.0f+16.0f));
					players[i]->setPosition(playerPos);
				}
			}
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::handleScore(bool player)
{
	int i;
	const TwoFloats& drawSize(getDrawingAreaSize());

	if(!player)
	{
		TwoFloats playerPos(players[0]->getPosition());
		ThreeFloats playerCol(players[0]->getColour());

		gameCounter.incrementCount(L"LeftScore");
		++leftType;
		//satCon += 1.0f;
		//saturation += 0.1f;
		shakeScreen();

		switch(leftType)
		{
			case Snake:
				particles.trigger(players[0]->getPosition(), players[0]->getColour());
				players[0] = shared_ptr<Player>(new SnakePlayer(playerPos,
																playerCol,
																gameCounter));
				setupPlayerKeys(0);

				addApple();
				break;
			case Football:
				particles.trigger(players[0]->getPosition(), players[0]->getColour());
				players[0] = shared_ptr<Player>(new FootballPlayer(playerPos, playerCol));
				setupPlayerKeys(0);
				rightGoals = true;
				break;
			case CaptureTheFlag:
				particles.trigger(players[0]->getPosition(), players[0]->getColour());
				players[0] = shared_ptr<Player>(new CaptureTheFlagPlayer(playerPos, playerCol));
				setupPlayerKeys(0);
				npcs.push_back(shared_ptr<Flag>(new Flag(TwoFloats(drawSize.x-64.0f, drawSize.y*0.5f),
														 ThreeFloats(0.53f, 0.45f, 1.0f))));
				break;
			case ZombieApocalypse:
				zombieCount[1] = 2;

				gameCounter.unregisterNotification(L"LeftScore", 1, this);
				gameCounter.unregisterNotification(L"RighttScore", 1, this);

				//Get rid of the associated flag.
				for(i=(npcs.size()-1);i>=0;--i)
				{
					if(npcs[i]->getName() == L"Flag")
					{
						shared_ptr<Flag> flag(dynamic_pointer_cast<Flag>(npcs[i]));

						if(flag->getColour() == ThreeFloats(0.53f, 0.45f, 1.0f))
						{
							npcs.erase(npcs.begin()+i);
							break;
						}
					}
				}
				break;
		}
	}
	else
	{
		TwoFloats playerPos(players[1]->getPosition());
		ThreeFloats playerCol(players[1]->getColour());
		
		gameCounter.incrementCount(L"RightScore");
		++rightType;
		//satCon += 1.0f;
		//saturation += 0.1f;
		shakeScreen();

		switch(rightType)
		{
			case Snake:
				particles.trigger(players[1]->getPosition(), players[1]->getColour());
				players[1] = shared_ptr<Player>(new SnakePlayer(playerPos,
																playerCol,
																gameCounter));
				setupPlayerKeys(1);
				addApple();
				break;
			case Football:
				particles.trigger(players[1]->getPosition(), players[1]->getColour());
				players[1] = shared_ptr<Player>(new FootballPlayer(playerPos, playerCol));
				setupPlayerKeys(1);
				leftGoals = true;
				break;
			case CaptureTheFlag:
				particles.trigger(players[1]->getPosition(), players[1]->getColour());
				players[1] = shared_ptr<Player>(new CaptureTheFlagPlayer(playerPos, playerCol));
				setupPlayerKeys(1);
				npcs.push_back(shared_ptr<Flag>(new Flag(TwoFloats(64.0f, drawSize.y*0.5f),
														 ThreeFloats(1.0f, 0.45f, 0.45f))));
				break;
			case ZombieApocalypse:
				zombieCount[0] = 2;

				gameCounter.unregisterNotification(L"LeftScore", 1, this);
				gameCounter.unregisterNotification(L"RighttScore", 1, this);

				//Get rid of the associated flag.
				for(i=(npcs.size()-1);i>=0;--i)
				{
					if(npcs[i]->getName() == L"Flag")
					{
						shared_ptr<Flag> flag(dynamic_pointer_cast<Flag>(npcs[i]));

						if(flag->getColour() == ThreeFloats(1.0f, 0.45f, 0.45f))
						{
							npcs.erase(npcs.begin()+i);
							break;
						}
					}
				}
				break;
		}
	}

	//Clean up any left over apples.
	if((leftType > Snake) && (rightType > Snake))
	{
		for(i=(npcs.size()-1);i>=0;--i)
		{
			if(npcs[i]->getName() == L"Apple")
				npcs.erase(npcs.begin()+i);
		}
	}

	ball->setPosition(drawSize*0.5f);
	ball->setOwner(0.0f);
	ball->resetSpeed(!player);

	if(!coloursSet)
	{
		ThreeFloats tempCol;

		players[0]->setColour(ThreeFloats(1.0f, 0.45f, 0.45f));
		players[1]->setColour(ThreeFloats(0.53f, 0.45f, 1.0f));

		tempCol.x = Random::getGlobal().getRandomFloat();
		tempCol.y = 0.4f;
		tempCol.z = 0.8f;
		tempCol = ColourHelpers::hslToRgb(tempCol);
		backgroundTween.reset(tempCol);

		tempCol.x = Random::getGlobal().getRandomFloat();
		tempCol.y = 0.66f;
		tempCol.z = 0.7f;
		tempCol = ColourHelpers::hslToRgb(tempCol);
		wallTween.reset(tempCol);
		
		coloursSet = true;
	}
}

//------------------------------------------------------------------------------
void GamePlayState::addApple()
{
	TwoFloats tempPos;
	shared_ptr<Apple> apple;
	const TwoFloats& drawSize(getDrawingAreaSize());

	tempPos.x = 48.0f + (Random::getGlobal().getRandomFloat() * (drawSize.x-96.0f));
	tempPos.y = 48.0f + (Random::getGlobal().getRandomFloat() * (drawSize.y-96.0f));

	apple = shared_ptr<Apple>(new Apple(tempPos));
	apple->setColour(appleColour);
	apple->setBackColour(starColour);
	npcs.push_back(apple);
}

//------------------------------------------------------------------------------
void GamePlayState::shakeScreen()
{
	if(gameOverColour == 0.0f)
		shakeVal += 0.5f;
}

//------------------------------------------------------------------------------
void GamePlayState::setupPlayerKeys(int index)
{
	if(upIsDown[index])
		players[index]->keyDown(L"Up");
	if(downIsDown[index])
		players[index]->keyDown(L"Down");
	if(leftIsDown[index])
		players[index]->keyDown(L"Left");
	if(rightIsDown[index])
		players[index]->keyDown(L"Right");
	if(joyLast[index].x != 0.0f)
		players[index]->joystick(L"Left", joyLast[index].x);
	if(joyLast[index].y != 0.0f)
		players[index]->joystick(L"Up", joyLast[index].y);
}
