//	SnakePlayer.cpp - The snake type player.
//  ----------------------------------------------------------------------------
//	This file is part of 'Minimal2PGame', a Ludum Dare entry.
//	Copyright (C) 2013  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "drawing/Drawer.h"
#include "SnakePlayer.h"
#include "GameCounter.h"
#include "Bullet.h"
#include "Random.h"
#include "Apple.h"
#include "Ball.h"

//------------------------------------------------------------------------------
SnakePlayer::SnakePlayer(const TwoFloats& pos,
						 const ThreeFloats& col,
						 GameCounter& counter):
Player(pos),
position(pos.x, pos.y-64.0f),
name(L"SnakePlayer"),
colour(col),
lastPos(position),
upIsDown(false),
downIsDown(false),
leftIsDown(false),
rightIsDown(false),
gameCounter(counter)
{
	int i;
	TwoFloats tailPos(position.x, position.y+128.0f);

	for(i=0;i<16;++i)
	{
		tail.push_back(tailPos);
		tailPos.y -= (128.0f/16.0f);

		tailScale.push_back(17.0f);
		tailCol.push_back(colour);
		tailAlpha.push_back(1.0f);
	}
	/*tail.push_back(position);
	tailScale.push_back(17.0f);
	tailCol.push_back(colour);
	tailAlpha.push_back(1.0f);
	tail.push_back(position);
	tailScale.push_back(17.0f);
	tailCol.push_back(colour);
	tailAlpha.push_back(1.0f);*/
}

//------------------------------------------------------------------------------
SnakePlayer::~SnakePlayer()
{

}

//------------------------------------------------------------------------------
void SnakePlayer::update(const TwoFloats& drawSize)
{
	const float speed(8.0f);

	position += direction * speed;
	if(position.x < 16.0f)
		position.x = 16.0f;
	else if(position.x > (drawSize.x-16.0f))
		position.x = drawSize.x-16.0f;
	if(position.y < (32.0f+16.0f))
		position.y = (32.0f+16.0f);
	else if(position.y > (drawSize.y-(32.0f+16.0f)))
		position.y = drawSize.y-(32.0f+16.0f);

	if(position != lastPos)
	{
		tail.erase(tail.begin());
		tail.push_back(position);

		lastPos = position;
	}
}

//------------------------------------------------------------------------------
void SnakePlayer::draw(shared_ptr<Drawer>& d)
{
	unsigned int i;

	d->drawConnectedLine(L"TailImage",
						 tail,
						 tailScale,
						 tailCol,
						 tailAlpha);

	for(i=0;i<tail.size();++i)
		d->drawImage(L"TriangleTailEnd", tail[i]-17.0f, 1.0f, colour);

	d->drawImage(L"TriangleTailEnd", position-17.0f, 1.0f, colour);
}

//------------------------------------------------------------------------------
void SnakePlayer::keyDown(const std::wstring& action)
{
	if(action == L"Up")
	{
		upIsDown = true;
		if(!downIsDown)
			direction.y = -1.0f;
		else
			direction.y = 0.0f;
	}
	else if(action == L"Down")
	{
		downIsDown = true;
		if(!upIsDown)
			direction.y = 1.0f;
		else
			direction.y = 0.0f;
	}
	else if(action == L"Left")
	{
		leftIsDown = true;
		if(!rightIsDown)
			direction.x = -1.0f;
		else
			direction.x = 0.0f;
	}
	else if(action == L"Right")
	{
		rightIsDown = true;
		if(!leftIsDown)
			direction.x = 1.0f;
		else
			direction.x = 0.0f;
	}
	direction = direction.normal();
}

//------------------------------------------------------------------------------
void SnakePlayer::keyUp(const std::wstring& action)
{
	if(action == L"Up")
	{
		upIsDown = false;
		if(!downIsDown)
			direction.y = 0.0f;
		else
			direction.y = 1.0f;
	}
	else if(action == L"Down")
	{
		downIsDown = false;
		if(!upIsDown)
			direction.y = 0.0f;
		else
			direction.y = -1.0f;
	}
	else if(action == L"Left")
	{
		leftIsDown = false;
		if(!rightIsDown)
			direction.x = 0.0f;
		else
			direction.x = 1.0f;
	}
	else if(action == L"Right")
	{
		rightIsDown = false;
		if(!leftIsDown)
			direction.x = 0.0f;
		else
			direction.x = -1.0f;
	}
	direction = direction.normal();
}

//------------------------------------------------------------------------------
void SnakePlayer::joystick(const std::wstring& action, float val)
{
	if((action == L"Up") || (action == L"Down"))
	{
		direction.y = val;
		if(fabsf(direction.y) < 0.15f)
			direction.y = 0.0f;
	}
	else if((action == L"Left") || (action == L"Right"))
	{
		direction.x = val;
		if(fabsf(direction.x) < 0.15f)
			direction.x = 0.0f;
	}
}

//------------------------------------------------------------------------------
void SnakePlayer::setPosition(const TwoFloats& pos)
{
	position = pos;
}

//------------------------------------------------------------------------------
void SnakePlayer::setDirection(const TwoFloats& dir)
{
	direction = dir;
}

//------------------------------------------------------------------------------
void SnakePlayer::collide(const shared_ptr<Collidable>& other)
{
	unsigned int i;
	TwoFloats tempPos;
	const float radius(getRadius());
	TwoFloats otherPos(other->getPosition());
	const float otherRad(other->getRadius());
	TwoFloats otherDir(other->getDirection());
	std::vector<TwoFloats> intersectingPoints;

	if(other->getName() == L"Apple")
	{
		if(otherPos.getDistance(position) < (otherRad+radius))
		{
			shared_ptr<Apple> apple(dynamic_pointer_cast<Apple>(other));

			if(apple)
				apple->setDead();

			for(i=0;i<8;++i)
			{
				tail.push_back(position);
				tailScale.push_back(17.0f);
				tailCol.push_back(colour);
				tailAlpha.push_back(1.0f);
			}

			if(colour == ThreeFloats(1.0f, 0.45f, 0.45f))
				gameCounter.incrementCount(L"LeftScore");
			else
				gameCounter.incrementCount(L"RightScore");
		}
	}
	else if((other->getName() != L"PongPlayer") &&
			(other->getName() != L"Flag"))
	{
		for(i=1;i<tail.size();++i)
		{
			if(otherPos.getDistance(tail[i]) < (otherRad+radius))
				intersectingPoints.push_back(tail[i]);
		}

		tempPos = 0.0f;
		if(intersectingPoints.size() > 0)
		{
			for(i=0;i<intersectingPoints.size();++i)
				tempPos += intersectingPoints[i];
			tempPos /= (float)intersectingPoints.size();

			if(other->getName() != L"Ball")
			{
				float offset;
				TwoFloats diff(tempPos - otherPos);
				float dist = diff.getMagnitude();
				float rad = getRadius() + other->getRadius();

				diff = diff.normal();
				offset = (rad - dist);
				otherPos -= diff * offset;

				other->setPosition(otherPos);
			}
			else if(other->getName() == L"Bullet")
			{
				shared_ptr<Bullet> bullet(dynamic_pointer_cast<Bullet>(other));

				bullet->setDead();
			}
			else
			{
				float reflectMag;
				TwoFloats mirrorVector((otherPos-tempPos).normal());
				shared_ptr<Ball> ball(dynamic_pointer_cast<Ball>(other));

				reflectMag = otherDir.getMagnitude();
				otherDir = otherDir.normal();
				otherDir = otherDir.mirror(mirrorVector).normal();
				otherDir *= reflectMag;

				other->setDirection(otherDir);
				other->setPosition(tempPos + (mirrorVector * (radius+otherRad)));

				ball->setOwner(0.0f);
			}
		}
	}
}

//------------------------------------------------------------------------------
void SnakePlayer::setColour(const ThreeFloats& col)
{
	unsigned int i;

	colour = col;

	for(i=0;i<tailCol.size();++i)
		tailCol[i] = colour;
}
