//	BrushStrokesSynth.h - Synth class based on my Brush Strokes VST plugin.
//  ----------------------------------------------------------------------------
//	This file is part of 'GameTemplate', template code for a game.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef BRUSHSTROKESSYNTH_H_
#define BRUSHSTROKESSYNTH_H_

#include "TwoFloats.h"
#include "Tweener.h"

///	Synth class based on my Brush Strokes VST plugin.
class BrushStrokesSynth
{
  public:
	///	Constructor.
	BrushStrokesSynth();
	///	Destructor.
	~BrushStrokesSynth();

	///	Returns the audio for this synth.
	/*!
		\param buffer Stereo (interleaved) buffer to write audio to.
		\param numSamples The number of samples in the current block (i.e.
		buffer will contain 2x numSamples.
	 */
	void getAudio(float *buffer, int numSamples);

	///	Sets the synth's frequency.
	void setFrequency(float val);
	///	Sets the synth's tables.
	/*!
		\param val This *MUST BE* an array of TableSize size.
	 */
	void setTables(const TwoFloats *val);

	///	Sets the synth's position in the stereo field.
	/*!
		\param val 0->1 (left->right).
	 */
	void setPan(float val);

	///	Turns the synth on or off.
	void setActive(bool val);
	///	Returns whether or not this synth is active.
	bool getActive() const {return active;};

	///	The size of the tables.
	enum
	{
		TableSize = 48
	};
  private:
	///	B-spline interpolation helper method.
	inline float bSplineInterp(float x0,
							   float x1,
							   float x2,
							   float x3,
							   float val)
	{
		// 4-point, 3rd-order B-spline (x-form)
		float ym1py1 = x0+x2;
		float c0 = (1.0f/6.0f) * ym1py1 + (2.0f/3.0f) * x1;
		float c1 = (1.0f/2.0f) * (x2-x0);
		float c2 = (1.0f/2.0f) * ym1py1 - x1;
		float c3 = (1.0f/2.0f) * (x1-x2) + (1.0f/6.0f) * (x3-x0);
		return ((c3*val+c2)*val+c1)*val+c0;
	};

	///	Whether or not this synth is currently active.
	bool active;
	///	Used to fade out the synth when setActive(false) is called.
	float fadeOut;
	///	Used to fade in the synth when setActive(true) is called.
	float fadeIn;

	///	The tables used to generate the sound.
	/*!
		These tables are only read from in the audio thread, and only written to
		in the main thread.

		It's TableSize+4 so we can use the extra 4 elements to loop round to the
		start of the buffer.

		We do without mutexes on the basis that we're not allocating/freeing any
		memory, and that it shouldn't matter too much if the tables change
		in the middle of a getAudio() call. (fingers crossed)
	 */
	TwoFloats tables[TableSize+4];
	///	The previous tables.
	/*!
		We interpolate between tables whenever setTables is called.
	 */
	TwoFloats prevTables[TableSize+4];
	///	Used to interpolate between the two tables.
	float tableInterp;

	///	The current frequency.
	float frequency;
	///	How much we increment position by for each sample.
	float increment;
	///	Used to glissando increment.
	Tweener<float> incrementTween;
	///	Used to prevent the frequency being updated too fast.
	int freqChangeCount;

	///	Our current position within the tables.
	float position;

	///	Our position in the stereo field.
	float pan;

	///	We filter any dc out of the output signal.
	TwoFloats dcSpeed;
	///	We filter any dc out of the output signal.
	TwoFloats dcPos;

	///	Temp.
	TwoFloats dcIn;
	///	Temp.
	TwoFloats dcOut;
	///	Temp.
	float dcCoeff;
};

#endif
